//
// Created by dawsh on 1/17/24.
//

#include <JGL/JGL.h>
#include <glad/glad.h>
#include <Color3.hpp>
#include <jlog/jlog.hpp>
#include <J3ML/Algorithm/Bezier.hpp>

GLfloat oldColor[4] = {0, 0, 0, 1};
GLfloat baseColor[4] = {1, 1, 1, 1};
bool inJ2D = false;
bool inJ3D = false;
bool wasTexture2DEnabled = false;
bool wasTextureCoordArrayEnabled = false;
bool wasDepthTestEnabled = false;
bool wasVertexArraysEnabled = false;
bool wasCullFaceEnabled = false;
bool wasBlendEnabled = false;
bool wasColorArrayEnabled = false;
GLint activeTextureUnit = 0;

namespace JGL {
    using namespace J3ML;
    Vector2 wS;

    void Update(const Vector2& window_size) {
        wS = window_size;
        glViewport(0, 0, wS.x, wS.y);
    }


    void J2D::Begin() {
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        glViewport(0, 0, wS.x, wS.y);
        glOrtho(0, wS.x, wS.y, 0, -1, 1);
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();

        //Get what the draw color was before we did anything.
        glGetFloatv(GL_CURRENT_COLOR, oldColor);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);

        glGetIntegerv(GL_ACTIVE_TEXTURE,& activeTextureUnit);
        activeTextureUnit = activeTextureUnit - GL_TEXTURE0;
        if (activeTextureUnit != 0)
            glActiveTexture(GL_TEXTURE0);

        wasDepthTestEnabled = false;
        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);

        wasVertexArraysEnabled = true;
        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnableClientState(GL_VERTEX_ARRAY);

        wasCullFaceEnabled = true;
        if (!glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = false,
            glEnable(GL_CULL_FACE),
            glCullFace(GL_BACK);

        wasBlendEnabled = true;
        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        wasTexture2DEnabled = true;
        if (!glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = false,
            glEnable(GL_TEXTURE_2D);

        wasTextureCoordArrayEnabled = true;
        if (!glIsEnabled(GL_TEXTURE_COORD_ARRAY))
            wasTextureCoordArrayEnabled = false,
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);

        wasColorArrayEnabled = false;
        if (glIsEnabled(GL_COLOR_ARRAY))
            wasColorArrayEnabled = true,
            glDisableClientState(GL_COLOR_ARRAY);

        if (!inJ3D)
            inJ2D = true;
        else { ERROR("Attempt to Begin J2D inside of J3D context.") }
    }

    void J2D::End() {
        //Change back to the previous projection (The 3D one in Re3D's case.)
        glPopMatrix();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisableClientState(GL_VERTEX_ARRAY);

        if (!wasCullFaceEnabled)
            glDisable(GL_CULL_FACE);

        if (!wasBlendEnabled)
            glDisable(GL_BLEND);

        if (!wasTexture2DEnabled)
            glDisable(GL_TEXTURE_2D);

        if (!wasTextureCoordArrayEnabled)
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

        if (wasColorArrayEnabled)
            glEnableClientState(GL_COLOR_ARRAY);

        //Select whatever texture_handle unit was selected before.
        glActiveTexture(GL_TEXTURE0 + activeTextureUnit);

        //Put the draw color back how it was before.
        glColor4f(oldColor[0], oldColor[1], oldColor[2], oldColor[3]);

        inJ2D = false;
    }

    void J2D::DrawSprite(const Texture& texture, const Vector2& pos, const Vector2& origin, const Vector2& scale, const Color4& color, Inversion inversion) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")


        const Vector2 size = texture.GetDimensions();

        std::array<Vector2, 4> textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};

        // TODO: Kind of a mess, refactor to be more sensible later.
        // Factors in scaling and origin correctly.
        // i.e. to render at 2x size, from the center, at coords XY, use {2, 2} scale, and {0.5, 0.5} offset.
        const Vector2 offset = origin * size;

        Vector2 pos2 = pos - offset * scale;
        Vector2 scaled_size = scale * size;
        Vector2 size2 = scaled_size;

        const Vector2 vertices[] = {
                pos2,                                 // Top-left vertex
                {pos2.x, pos2.y + size2.y},            // Bottom-left
                {pos2.x + size2.x, pos2.y + size2.y},   // Bottom-right
                {pos2.x + size2.x, pos2.y}             // Top-right
        };

        if (inversion&  Inversion::Vertical)
            textureCoordinates = {Vector2(0, 1), Vector2(0, 0), Vector2(1, 0), Vector2(1, 1)};
        if (inversion&  Inversion::Horizontal)
            textureCoordinates = {Vector2(1, 0), Vector2(1, 1), Vector2(0, 1), Vector2(0, 0)};
        if ((inversion& Inversion::Horizontal) && (inversion& Inversion::Vertical))
            textureCoordinates = {Vector2(1, 1), Vector2(1, 0), Vector2(0, 0), Vector2(0, 1)};

        glColor4ubv(color.ptr());
        glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());
        glDrawArrays(GL_QUADS, 0, 4);
        glBindTexture(GL_TEXTURE_2D, 0);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }


    void J2D::DrawSprite(const Texture& texture, float positionX, float positionY, float originX, float originY, float scaleX, float scaleY, const Color4& color, Inversion inversion)
    {
        DrawSprite(texture,
                   {positionX, positionX},
                   {originX, originY},
                   {scaleX, scaleY},
                   color,
                   inversion);
    }

    void J2D::DrawPartialSprite(const Texture& texture, const Vector2& position, const Vector2& sub_texture_position, const Vector2& sub_texture_size, const Vector2& origin, const Vector2& scale, const Color4& color, Inversion inversion) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.");

        const Vector2 textureSize = texture.GetDimensions();

        // Calculate texture coordinates (correctly relative to the whole texture)
        std::array<GLfloat, 8> textureCoordinates = {
                sub_texture_position.x / textureSize.x,
                sub_texture_position.y / textureSize.y,
                sub_texture_position.x / textureSize.x,
                (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                sub_texture_position.y / textureSize.y
        };

        if (inversion & Inversion::Vertical) {
            std::swap(textureCoordinates[1], textureCoordinates[3]);
            std::swap(textureCoordinates[5], textureCoordinates[7]);
        }
        if (inversion & Inversion::Horizontal) {
            std::swap(textureCoordinates[0], textureCoordinates[6]);
            std::swap(textureCoordinates[2], textureCoordinates[4]);
        }

        const Vector2 offset = origin * sub_texture_size;

        Vector2 pos2 = position - offset * scale;
        Vector2 scaled_size = scale * sub_texture_size;
        Vector2 size2 = scaled_size;

        const Vector2 vertices[] = {
                pos2,                                // Top-left
                {pos2.x, pos2.y + size2.y},          // Bottom-left
                {pos2.x + size2.x, pos2.y + size2.y},// Bottom-right
                {pos2.x + size2.x, pos2.y}           // Top-right
        };

        glColor4ubv(color.ptr());
        glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glTexCoordPointer(2, GL_FLOAT, 0, textureCoordinates.data());
        glDrawArrays(GL_QUADS, 0, 4);
        glBindTexture(GL_TEXTURE_2D, 0);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::DrawPartialSprite(const JGL::Texture& texture, float positionX, float positionY, float sub_texture_positionX, float sub_texture_positionY, unsigned int sub_texture_sizeX,
                                unsigned int sub_texture_sizeY, float originX,float originY, float scaleX, float scaleY, const Color4& color,JGL::Inversion inversion) {

        J2D::DrawPartialSprite(texture, {positionX, positionY}, {sub_texture_positionX, sub_texture_positionY},
                               {(float) sub_texture_sizeX, (float) sub_texture_sizeY}, {originX, originY}, {scaleX, scaleY}, color, inversion);
    }

    void J2D::FillQuad(const Color4& color, const Vector2& v1, const Vector2& v2, const Vector2& v3, const Vector2& v4) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {v1, v2, v3, v4};
        glColor4f(color.RedChannelNormalized(),
                  color.GreenChannelNormalized(),
                  color.BlueChannelNormalized(),
                  color.AlphaChannelNormalized());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_QUADS, 0, 4);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::FillQuad(const Color3& color, const Vector2& v1, const Vector2& v2, const Vector2& v3, const Vector2& v4) {
        J2D::FillQuad(Color4(color), v1, v2, v3, v4);
    }

    void J2D::OutlineQuad(const Color4& color, const Vector2& v1, const Vector2& v2, const Vector2& v3, const Vector2& v4, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {v1, v2, v3, v4};
        glLineWidth(thickness);
        glColor4f(color.RedChannelNormalized(),
                  color.GreenChannelNormalized(),
                  color.BlueChannelNormalized(),
                  color.AlphaChannelNormalized());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 4);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::OutlineQuad(const Color3& color, const Vector2& v1, const Vector2& v2, const Vector2& v3, const Vector2& v4, float thickness) {
        J2D::OutlineQuad(Color4(color), v1, v2, v3, v4);
    }



    void J2D::FillRect(const Color4& color, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        glColor4f(color.RedChannelNormalized(),
                  color.GreenChannelNormalized(),
                  color.BlueChannelNormalized(),
                  color.AlphaChannelNormalized());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_QUADS, 0, 4);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::FillRect(const Color3& color, const Vector2& pos, const Vector2& size) {
        J2D::FillRect({color.r, color.g, color.b, 255}, pos, size);
    }

    void J2D::FillGradientRect(const Color4& color1, const Color4& color2, const Gradient& gradient, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        std::vector<GLfloat> colors = {};

        if (gradient == Gradient::Horizontal)
            colors = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f, color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,
                     color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f, color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f};

        else if (gradient == Gradient::Vertical)
            colors = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f, color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f,
                     color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f, color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f};

        else if (gradient == Gradient::DiagonalBottomLeft)
            colors = {(color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f, (color1.b + color2.b) / 2.f / 255.f, (color1.a + color2.a) / 2.f / 255.f,
                    color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,(color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f,
                    (color1.b + color2.b) / 2.f / 255.f, (color1.a + color2.a) / 2.f / 255.f, color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f};

        else if (gradient == Gradient::DiagonalTopLeft)
                    colors = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,(color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f,
                              (color1.b + color2.b) / 2.f / 255.f, (color1.a + color2.a) / 2.f / 255.f,color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f,
                              (color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f, (color1.b + color2.b) / 2.f / 255.f,(color1.a + color2.a) / 2.f / 255.f};

        glEnableClientState(GL_COLOR_ARRAY);
            glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
            glColorPointer(4, GL_FLOAT, 0, colors.data());
            glDrawArrays(GL_QUADS, 0, 4);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::FillGradientRect(const Color3& color1, const Color3& color2, const Gradient& gradient, const Vector2& pos, const Vector2& size) {
        J2D::FillGradientRect({color1.r, color1.g, color1.b, 255}, {color2.r, color2.g, color2.b, 255}, gradient, pos, size);
    }

    void J2D::FillRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        J2D::FillRect(color, {pos.x + radius, pos.y}, {size.x - 2 * radius, size.y});
        J2D::FillRect(color, {pos.x, pos.y + radius}, {size.x, size.y - 2 * radius});

        J2D::FillCircle(color, {pos.x + radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + radius, pos.y + size.y - radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + size.y - radius}, radius, subdivisions);
    }

    void J2D::FillRoundedRect(const Color3& color, const J3ML::LinearAlgebra::Vector2& pos, const J3ML::LinearAlgebra::Vector2& size, float radius, unsigned int subdivisions) {
        J2D::FillRoundedRect({color.r, color.g, color.b, 255}, pos, size, radius, subdivisions);
    }

    void J2D::OutlineRect(const Color4& color, const Vector2& pos, const Vector2& size, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};

        glLineWidth(thickness);
        glColor4f(color.RedChannelNormalized(),
                  color.GreenChannelNormalized(),
                  color.BlueChannelNormalized(),
                  color.AlphaChannelNormalized());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 4);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::OutlineRect(const Color3& color, const Vector2& pos, const Vector2& size, float thickness) {
        J2D::OutlineRect({color.r, color.g, color.b, 255}, pos, size, thickness);
    }

    void J2D::DrawLine(const Color4& color, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.");

        Vector2 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4f(color.RedChannelNormalized(),
                  color.GreenChannelNormalized(),
                  color.BlueChannelNormalized(),
                  color.AlphaChannelNormalized());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::DrawLine(const Color3& color, const Vector2& A, const Vector2& B, float thickness) {
        J2D::DrawLine({color.r, color.g, color.b, 255}, A, B, thickness);
    }

    void J2D::DrawLine(const Color4& color, float x, float y, float w, float h, float thickness) {
        J2D::DrawLine(color, {x, y}, {w, h}, thickness);
    }

    void J2D::DrawLine(const Color3& color, float x, float y, float w, float h, float thickness) {
        J2D::DrawLine({color.r, color.g, color.b, 255}, x, y, w, h, thickness);
    }

    void J2D::DrawGradientLine(const Color4& color1, const Color4& color2, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.");

        Vector2 vertices[] = {A, B};
        GLfloat colors[8] = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,
                             color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f};

        glEnableClientState(GL_COLOR_ARRAY);
            glLineWidth(thickness);
            glColorPointer(4,GL_FLOAT,sizeof(GL_FLOAT) * 4, colors);
            glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
            glDrawArrays(GL_LINES, 0, 2);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::DrawGradientLine(const Color3& color1, const Color3& color2, const Vector2& A, const Vector2& B, float thickness) {
        J2D::DrawGradientLine({color1.r, color1.g, color1.b, 255}, {color2.r, color2.g, color2.b, 255}, A, B, thickness);
    }
    void DrawGradientLine(const Color4& color1, const Color4& color2, float x, float y, float w, float h, float thickness) {
        J2D::DrawGradientLine(color1, color2, {x, y}, {w, h}, thickness);
    }

    void DrawGradientLine(const Color3& color1, const Color3& color2, float x, float y, float w, float h, float thickness) {
        J2D::DrawGradientLine({color1.r, color1.g, color1.b, 255}, {color2.r, color2.g, color2.b, 255}, {x, y}, {w, h}, thickness);
    }

    void J2D::DrawPoint(const Color4& color, const Vector2& coordinates, float radius) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.");

        Vector2 vertices[] = {coordinates};

        glPointSize(radius);
        glColor4f(color.RedChannelNormalized(),
                  color.GreenChannelNormalized(),
                  color.BlueChannelNormalized(),
                  color.AlphaChannelNormalized());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_POINTS, 0, 1);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::DrawPoint(const Color3& color, const Vector2& coordinates, float radius) {
        J2D::DrawPoint({color.r, color.g, color.b, 255}, coordinates);
    }

    void J2D::DrawPoint(const Color4& color, float x, float y, float radius) {
        DrawPoint(color, {x, y});
    }

    void J2D::DrawPoint(const Color3& color, float x, float y, float radius) {
        DrawPoint({color.r, color.g, color.b, 255}, {x, y});
    }

    void J2D::OutlineCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        float step = (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices{};
        GLfloat angle, x, y;

        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
            x = radius * std::sin(angle) + center.x;
            y = radius * std::cos(angle) + center.y;
            vertices.emplace_back(x,y);
        }

        glLineWidth(thickness);
        glColor4f(color.RedChannelNormalized(),
                  color.GreenChannelNormalized(),
                  color.BlueChannelNormalized(),
                  color.AlphaChannelNormalized());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_LINE_LOOP, 0, vertices.size());
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::OutlineCircle(const Color3& color, const Vector2& center, float radius, unsigned int subdivisions, float thickness) {
        J2D::OutlineCircle({color.r, color.g, color.b, 255}, center, radius, subdivisions, thickness);
    }

    void J2D::FillCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        GLfloat angle, x, y;
        float step =  (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices{};

        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step)
                x = radius * sin(angle) + center.x,
                y = radius * cos(angle) + center.y,
                vertices.push_back({x, y});

        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_TRIANGLE_FAN, 0, vertices.size());
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::FillCircle(const Color3& color, const Vector2& center, float radius, unsigned int subdivisions) {
        J2D::FillCircle({color.r, color.g, color.b, 255}, center, radius, subdivisions);
    }

    void J2D::OutlineTriangle(const Color4& color, const Triangle2D& tri, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glLineWidth(thickness);
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 3);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::OutlineTriangle(const Color3& color, const Triangle2D& tri, float thickness) {
        J2D::OutlineTriangle({color.r, color.g, color.b, 255}, tri, thickness);
    }

    void J2D::FillTriangle(const Color4& color, const Triangle2D& tri) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::FillGradientTriangle(const Color4& a_color, const Color4& b_color, const Color4& c_color, const Triangle2D& tri) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};
        GLfloat colors[] = {a_color.r / 255.f, a_color.g / 255.f, a_color.b / 255.f, a_color.a / 255.f,b_color.r / 255.f,
                            b_color.g / 255.f, b_color.b / 255.f, b_color.a / 255.f,c_color.r / 255.f, c_color.g / 255.f, c_color.b / 255.f,
                            c_color.a / 255.f };

        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(4, GL_FLOAT, sizeof(Color4), colors);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J2D::FillGradientTriangle(const Color3& a_color, const Color3& b_color, const Color3& c_color, const Triangle2D& tri) {
        J2D::FillGradientTriangle(Color4(a_color), Color4(b_color), Color4(c_color), tri);
    }


    void J2D::FillTriangle(const Color3& color, const Triangle2D& tri) {
        J2D::FillTriangle({color.r, color.g, color.b, 255}, tri);
    }



    void J2D::DrawCubicBezierCurve(const Color4 &color, const J3ML::LinearAlgebra::Vector2 &controlA,
                                   const J3ML::LinearAlgebra::Vector2 &pointA,
                                   const J3ML::LinearAlgebra::Vector2 &pointB,
                                   const J3ML::LinearAlgebra::Vector2 &controlB, int subdivisions, float thickness) {


        Vector2 last = controlA;
        Vector2 first = controlB;
        for (int i = 0; i < subdivisions; ++i)
        {
            float alpha = (float)i / (float)subdivisions;
            Vector2 step = J3ML::Algorithm::Bezier(alpha, controlA, pointA, pointB, controlB);
            DrawLine(color, last, step, thickness);
            last = step;
        }

        // Have to manually draw the last segment of the curve.
        DrawLine(color, last, first, thickness);

        // Display control points
        DrawPoint(Colors::Red, controlA, 2.f);
        DrawPoint(Colors::Red, controlB, 2.f);
        DrawPoint(Colors::Reds::Salmon, pointA, 2.f);
        DrawPoint(Colors::Reds::Salmon, pointB, 2.f);

    }

    //The 3D projection.
    std::vector<GLfloat> perspective(float fov, float aspect, float nearPlane, float farPlane) {
        std::vector<float> result(16);
        float f = 1.0f / tan(fov * 0.5f * Math::Pi / 180.0f);
        result[0] = f / aspect;
        result[5] = f;
        result[10] = (farPlane + nearPlane) / (nearPlane - farPlane);
        result[11] = -1.0f;
        result[14] = (2.0f * farPlane * nearPlane) / (nearPlane - farPlane);
        return result;
    }

    bool j3d_initialized = false;
    float j3d_far_plane = 0;
    float j3d_fov = 0;

    void J3D::Init(const J3ML::LinearAlgebra::Vector2& window_size, float fov, float far_plane) {
        wS = window_size;
        j3d_far_plane = far_plane;
        j3d_fov = fov;
        j3d_initialized = true;
    }

    void J3D::ChangeFOV(float fov) {
        j3d_fov = fov;
    }

    void J3D::ChangeFarPlane(float far_plane) {
        j3d_far_plane = far_plane;
    }

    void J3D::Begin() {
        if (!j3d_initialized)
            throw std::runtime_error("You have to run J3D::Init before rendering 3D elements.");

        auto aspect = (float) wS.x / (float) wS.y;
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glMultMatrixf(perspective(j3d_fov, aspect, 0.001, j3d_far_plane).data());
        glMatrixMode(GL_MODELVIEW);

        //Get what the draw color was before we did anything.
        glGetFloatv(GL_CURRENT_COLOR, oldColor);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);

        wasDepthTestEnabled = false;
        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);

        wasVertexArraysEnabled = false;
        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnableClientState(GL_VERTEX_ARRAY);

        wasTexture2DEnabled = true;
        if (!glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = false,
            glEnable(GL_TEXTURE_2D);

        // TODO: implement bool drawBackface as DrawString parameter.
        wasCullFaceEnabled = false;
        if (glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = true,
            glDisable(GL_CULL_FACE);

        wasBlendEnabled = true;
        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);



        if (!inJ2D)
            inJ3D = true;
        else { ERROR("Attempt to Begin J3D inside of J2D context.")}
    }

    void J3D::End() {
        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisableClientState(GL_VERTEX_ARRAY);

        if (wasTexture2DEnabled)
            glDisable(GL_TEXTURE_2D);

        if (!wasBlendEnabled)
            glDisable(GL_BLEND);

        if (wasCullFaceEnabled)
            glEnable(GL_CULL_FACE);

        //Put the draw color back how it was before.
        glColor4f(oldColor[0], oldColor[1], oldColor[2], oldColor[3]);

        inJ3D = false;
    }

    void J3D::DrawLine(const Color4& color, const Vector3& A, const Vector3& B, float thickness) {
        if (!inJ3D)
            ERROR("Attempt to Render J3D element before J3D begin.")

        Vector3 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glColor4f(baseColor[0], baseColor[1], baseColor[2], baseColor[3]);
    }

    void J3D::DrawLine(const Color3& color, const Vector3& A, const Vector3& B, float thickness) {
        J3D::DrawLine({color.r, color.g, color.b, 255}, A, B, thickness);
    }
}
