#include <JGL/Texture.h>
#include <iostream>

using namespace ReTexture;

namespace JGL
{
    Texture::Texture(const std::string &file, const ReTexture::TextureFlag &flags, TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode)
    {
        auto *t = new ReTexture::SoftwareTexture(file, flags);


        load(t, {(float) t->getWidth(), (float) t->getHeight()}, t->getTextureFormat(), filtering_mode,
             wrapping_mode);
        texture_flags = flags;

        delete t;
    }

    Texture::Texture(const std::string &file, TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode) {
        auto *t = new SoftwareTexture(file);

        load(t, {(float) t->getWidth(), (float) t->getHeight()}, t->getTextureFormat(), filtering_mode,
             wrapping_mode);
        texture_flags = TextureFlag::NONE;

        delete t;
    }

    void Texture::load(SoftwareTexture *software_texture, const Vector2 &size, const TextureFormat &format,
                           TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode) {
            glGenTextures(1, &texture_handle);
            glBindTexture(GL_TEXTURE_2D, texture_handle);

            if (format == TextureFormat::RGBA)
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (int) size.x, (int) size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                             software_texture->pixelData.data());

            else if (format == TextureFormat::RGB)
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, (int) size.x, (int) size.y, 0, GL_RGB, GL_UNSIGNED_BYTE,
                             software_texture->pixelData.data());

            if (wrapping_mode == TextureWrappingMode::CLAMP_TO_EDGE)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE),
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

            else if (wrapping_mode == TextureWrappingMode::REPEAT)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT),
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

            else if (wrapping_mode == TextureWrappingMode::MIRRORED_REPEAT)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT),
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);

            else if (wrapping_mode == TextureWrappingMode::CLAMP_TO_BORDER)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER),
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
            if (filtering_mode == TextureFilteringMode::NEAREST)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST),
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

            else if (filtering_mode == TextureFilteringMode::BILINEAR)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR),
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

            else if (filtering_mode == TextureFilteringMode::MIPMAP_NEAREST ||
                     filtering_mode == TextureFilteringMode::MIPMAP_BILINEAR ||
                     filtering_mode == TextureFilteringMode::MIPMAP_TRILINEAR) {
                //3 mipmap levels.
                auto *m1 = new SoftwareTexture(software_texture->downscale(2));
                auto *m2 = new SoftwareTexture(software_texture->downscale(4));
                auto *m3 = new SoftwareTexture(software_texture->downscale(8));

                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 3);

                if (format == TextureFormat::RGBA) {
                    glTexImage2D(GL_TEXTURE_2D, 1, GL_RGBA, m1->getWidth(), m1->getHeight(), 0, GL_RGBA,
                                 GL_UNSIGNED_BYTE, m1->pixelData.data());
                    glTexImage2D(GL_TEXTURE_2D, 2, GL_RGBA, m2->getWidth(), m2->getHeight(), 0, GL_RGBA,
                                 GL_UNSIGNED_BYTE, m2->pixelData.data());
                    glTexImage2D(GL_TEXTURE_2D, 3, GL_RGBA, m3->getWidth(), m3->getHeight(), 0, GL_RGBA,
                                 GL_UNSIGNED_BYTE, m3->pixelData.data());
                } else if (format == TextureFormat::RGB) {
                    glTexImage2D(GL_TEXTURE_2D, 1, GL_RGB, m1->getWidth(), m1->getHeight(), 0, GL_RGB, GL_UNSIGNED_BYTE,
                                 m1->pixelData.data());
                    glTexImage2D(GL_TEXTURE_2D, 2, GL_RGB, m2->getWidth(), m2->getHeight(), 0, GL_RGB, GL_UNSIGNED_BYTE,
                                 m2->pixelData.data());
                    glTexImage2D(GL_TEXTURE_2D, 3, GL_RGB, m3->getWidth(), m3->getHeight(), 0, GL_RGB, GL_UNSIGNED_BYTE,
                                 m3->pixelData.data());
                }

                if (filtering_mode == TextureFilteringMode::MIPMAP_NEAREST)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST),
                            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

                else if (filtering_mode == TextureFilteringMode::MIPMAP_BILINEAR)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST),
                            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

                else if (filtering_mode == TextureFilteringMode::MIPMAP_TRILINEAR)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR),
                            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

                delete m1;
                delete m2;
                delete m3;
            }
            glBindTexture(GL_TEXTURE_2D, 0);

            texture_size = size;
            texture_format = format;
            texture_filtering_mode = filtering_mode;
    }

    std::vector<Color4> JGL::Texture::GetPixelData() const {
            std::vector<Color4> result((size_t) (texture_size.x * texture_size.y));
            glBindTexture(GL_TEXTURE_2D, texture_handle);

            if (texture_format == TextureFormat::RGBA) {
                glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, result.data());
                return result;
            }

            //if RGB
            std::vector<Color3> color3((size_t) (texture_size.x * texture_size.y));

            glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, color3.data());

            for (const auto &c: color3)
                result.emplace_back(c);
            return result;
        }

        void Texture::Erase() {
            if (texture_handle != 0)
                glDeleteTextures(1, &texture_handle);
        }

        GLuint Texture::GetGLTextureHandle() const {
            return texture_handle;
        }

        Vector2 Texture::GetDimensions() const {
            return texture_size;
        }

        TextureFlag Texture::GetFlags() const {
            return texture_flags;
        }

        TextureFormat Texture::GetFormat() const {
            return texture_format;
        }

    TextureFilteringMode Texture::GetFilteringMode() const {
        return texture_filtering_mode;
    }

    void Texture::SetTextureHandle(GLuint handle) {
        texture_handle = handle;
    }
}